"""
智能输液监控系统（MicroPython版）
功能：实时监测输液进度、计算滴速、低液量报警、数据上报
硬件依赖：行空板K10、红外传感器、有源蜂鸣器
"""

from educore import pin, wifi, mqttclient, oled, speaker, uuid
import time
import random
import sys
import utime

# --------------------------
# 配置中心（需根据实际硬件调整）
# --------------------------
class AppConfig:
    """应用程序全局配置"""
    
    # WiFi配置
    WIFI_SSID = "GOCLOUD"          # 无线网络名称
    WIFI_PASSWORD = "L2100Hbb"     # WiFi密码
    WIFI_TIMEOUT_MS = 50000        # 连接超时（毫秒）
    WIFI_MAX_RETRIES = 3           # 最大重试次数
    
    # MQTT配置
    MQTT_SERVER = "192.168.123.16"       # 消息代理地址
    MQTT_PORT = 1883                     # 服务端口
    MQTT_USER = "siot"                   # 认证用户名
    MQTT_PASSWORD = "dfrobot"            # 认证密码
    MQTT_MAX_RETRIES = 3                 # 连接重试次数
    ALARM_TOPIC = "siot/infusion_alarm"  # 输液报警主题
    START_TOPIC = "siot/infusion_start"  # 输液开始主题
    END_TOPIC = "siot/infusion_end"      # 输液结束主题
    
    # 硬件参数
    HARDWARE_DEBOUNCE_MS = 50      # 传感器消抖时间（毫秒）
    DISPLAY_REFRESH_MS = 1000      # 屏幕刷新间隔（毫秒）
    BUZZER_FREQ = 2000             # 蜂鸣器频率（Hz）
    BUZZER_DURATION_MS = 500       # 蜂鸣持续时间（毫秒）
    SENSOR_PIN = 0                 # 滴液传感器GPIO引脚
    SPEAKER_PIN = 1                # 蜂鸣器控制引脚
    
    # 安全参数（成人标准）
    SAFETY_MIN_DROP_RATE = 40      # 最低安全滴速（滴/分钟）
    SAFETY_MAX_DROP_RATE = 60      # 最高安全滴速（滴/分钟）

# --------------------------
# 数据模型
# --------------------------
class InfusionMetrics:
    """输液过程指标数据容器
    
    属性：
    - drop_rate: 当前滴速（滴/分钟）
    - elapsed_time: 已输注时间（HH:MM:SS）
    - remaining_volume: 剩余液量（毫升）
    - remaining_time: 预计剩余时间（HH:MM:SS）
    """
    def __init__(self, drop_rate, elapsed_time, remaining_volume, remaining_time):
        self.drop_rate = drop_rate
        self.elapsed_time = elapsed_time
        self.remaining_volume = remaining_volume
        self.remaining_time = remaining_time

# --------------------------
# 工具函数
# --------------------------
def format_duration(minutes):
    """将分钟数格式化为HH:MM:SS
    
    参数：
    minutes -- 总分钟数（浮点数）
    
    返回：
    格式化后的时间字符串，如 01:23:45
    """
    total_sec = int(minutes * 60)
    hours = total_sec // 3600
    minutes = (total_sec % 3600) // 60
    seconds = total_sec % 60
    return "{:02d}:{:02d}:{:02d}".format(hours, minutes, seconds)

def clamp(value, min_val, max_val):
    """数值范围约束函数
    
    参数：
    value -- 原始值
    min_val -- 允许的最小值
    max_val -- 允许的最大值
    
    返回：
    被限制在[min_val, max_val]范围内的值
    """
    return max(min_val, min(value, max_val))

def datetime():
    """获取当前日期与时间
    返回：
    格式yyyy-mm-dd hh:mm:ss，如2025-02-06 12:28:15
    """
    # 获取当前时间戳
    timestamp = utime.localtime()
    # 格式化时间
    date_time = "{:04d}-{:02d}-{:02d} {:02d}:{:02d}:{:02d}".format(
        timestamp[0], timestamp[1], timestamp[2],
        timestamp[3], timestamp[4], timestamp[5]
    )
    return date_time

# --------------------------
# 核心算法
# --------------------------
class InfusionCalculator:
    """输液指标计算引擎"""
    
    @staticmethod
    def calculate(current_time, start_time, drop_count, total_ml, drop_factor, cfg):
        """执行所有指标计算
        
        参数：
        current_time -- 当前时间（毫秒ticks）
        start_time -- 输液开始时间（毫秒ticks）
        drop_count -- 累计滴数
        total_ml -- 输液总量（毫升）
        drop_factor -- 滴系数（滴/毫升）
        cfg -- 配置对象
        
        返回：
        InfusionMetrics对象包含计算结果
        """
        # 计算经过时间（防止除零）
        elapsed_ticks = max(time.ticks_diff(current_time, start_time), 1)
        elapsed_min = elapsed_ticks / 60000  # 转换为分钟
        
        # 有效滴数处理（防止负数）
        valid_drops = max(drop_count, 0)
        
        # 计算原始滴速并约束到安全范围
        raw_rate = valid_drops / elapsed_min
        safe_rate = clamp(raw_rate, cfg.SAFETY_MIN_DROP_RATE, cfg.SAFETY_MAX_DROP_RATE)
        
        # 计算液量
        infused_ml = min(valid_drops / drop_factor, total_ml)
        remaining_ml = max(total_ml - infused_ml, 0)
        
        # 估算剩余时间（基于安全滴速）
        remain_min = (remaining_ml * drop_factor / safe_rate) if safe_rate > 0 else 0
        
        return InfusionMetrics(
            safe_rate,
            format_duration(elapsed_min),
            remaining_ml,
            format_duration(remain_min)
        )

# --------------------------
# 主监控系统
# --------------------------
class InfusionMonitor:
    """输液监控主控制器"""
    
    def __init__(self, cfg, bed_name, total_volume, drop_factor, alarm_threshold):
        """初始化监控实例
        
        参数：
        cfg -- AppConfig配置对象
        bed_name -- 床位标识字符串
        total_volume -- 总液量（毫升）
        drop_factor -- 滴系数（滴/毫升）
        alarm_threshold -- 低液量报警阈值（毫升）
        """
        # 参数有效性检查
        if total_volume <= 0 or drop_factor <= 0 or alarm_threshold <= 0:
            raise ValueError("参数必须为正数")
        
        self.cfg = cfg
        self.bed_name = bed_name
        self.total_volume = total_volume
        self.drop_factor = drop_factor
        self.alarm_threshold = alarm_threshold
        self.uuid = uuid()
        
        # 初始化运行状态
        self._reset_state()
        
        # 初始化硬件接口
        self._init_hardware()

    def _reset_state(self):
        """重置所有运行时状态"""
        self.drop_count = 0            # 累计滴数
        self.start_time = time.ticks_ms()  # 输液开始时间
        self.last_sensor_state = None  # 传感器上次状态
        self.last_debounce_time = 0    # 上次消抖时间
        self.has_alarmed = False       # 报警状态标记
        self.is_running = True         # 运行状态标志

    def _init_hardware(self):
        """初始化硬件设备"""
        # 滴液传感器（上拉输入模式）
        self.drop_sensor = pin(self.cfg.SENSOR_PIN)
        
        # OLED显示屏初始化
        self.display = oled
        
        # 蜂鸣器初始化
        self.buzzer = speaker(self.cfg.SPEAKER_PIN)
        self.buzzer.stop()

    def connect_network(self):
        """建立网络连接（WiFi+MQTT）"""
        self._connect_wifi()
        self._connect_mqtt()

    def _connect_wifi(self):
        """连接WiFi网络（带重试机制）"""
        for attempt in range(1, self.cfg.WIFI_MAX_RETRIES + 1):
            try:
                wifi.connect(
                    ssid=self.cfg.WIFI_SSID,
                    psd=self.cfg.WIFI_PASSWORD,
                    timeout=self.cfg.WIFI_TIMEOUT_MS
                )
                if wifi.status():
                    print("[网络] WiFi连接成功")
                    return
            except Exception as e:
                print(f"[网络] WiFi连接异常（尝试{attempt}/{self.cfg.WIFI_MAX_RETRIES}）: {str(e)}")
        raise RuntimeError("WiFi连接失败")

    def _connect_mqtt(self):
        """连接MQTT代理（带重试机制）"""
        for attempt in range(1, self.cfg.MQTT_MAX_RETRIES + 1):
            try:
                mqttclient.connect(
                    server=self.cfg.MQTT_SERVER,
                    port=self.cfg.MQTT_PORT,
                    client_id=self.bed_name,
                    user=self.cfg.MQTT_USER,
                    psd=self.cfg.MQTT_PASSWORD
                )
                if mqttclient.connected():
                    print("[网络] MQTT连接成功")
                    return
            except Exception as e:
                print(f"[网络] MQTT连接异常（尝试{attempt}/{self.cfg.MQTT_MAX_RETRIES}）: {str(e)}")
        raise RuntimeError("MQTT连接失败")

    def _handle_sensor(self, current_time):
        """处理滴液传感器信号（带消抖）"""
        current_state = self.drop_sensor.read_digital()
        
        # 状态变化检测
        if current_state != self.last_sensor_state:
            # 消抖时间检查
            if time.ticks_diff(current_time, self.last_debounce_time) > self.cfg.HARDWARE_DEBOUNCE_MS:
                # 下降沿触发计数（传感器输出低电平）
                if current_state == 0:
                    max_drops = self.total_volume * self.drop_factor
                    if self.drop_count < max_drops:
                        self.drop_count += 1
                        print(f"[传感器] 滴数更新：{self.drop_count}")
                
                # 更新状态记录
                self.last_sensor_state = current_state
                self.last_debounce_time = current_time

    def _update_interface(self, current_time):
        """更新用户界面并返回计算结果"""
        # 计算各项指标
        metrics = InfusionCalculator.calculate(
            current_time, self.start_time, self.drop_count,
            self.total_volume, self.drop_factor, self.cfg
        )       
        # 清屏并绘制界面
        self.display.clear()
        self._render_header()
        self._render_metrics(metrics)       
        # 液量状态检查
        if metrics.remaining_volume <= 0:
            self.__recode_end()       #输液结束记录
            self._shutdown()
        elif metrics.remaining_volume < self.alarm_threshold:
            self._trigger_alarm(metrics.remaining_volume)        
        return metrics

    def _render_header(self):
        """绘制界面固定标题部分"""
        self.display.print("Bed {}".format(self.bed_name), 80, 20)  # 床位号
        self.display.print("Factor: {}/mL".format(self.drop_factor), 20, 50)  # 滴系数

    def _render_metrics(self, metrics):
        """绘制动态监测指标"""
        y_pos = [90, 140, 190, 240]  # 四行数据的Y坐标
        self.display.print("Rate: {:.1f}/min".format(metrics.drop_rate), 20, y_pos[0])
        self.display.print("Elapsed: {}".format(metrics.elapsed_time), 20, y_pos[1])
        self.display.print("Volume: {:.1f}mL".format(metrics.remaining_volume), 20, y_pos[2])
        self.display.print("Remain: {}".format(metrics.remaining_time), 20, y_pos[3])

    def _trigger_alarm(self, remaining_volume):
        """触发低液量报警流程
        报警信息格式：SIM-1|low volume:0.95mL
        """
        volume = remaining_volume
        # 视觉报警
        self.display.print("! LOW VOLUME !", 30, 280)
        # 网络报警（仅首次触发）
        if not self.has_alarmed:
            try:
                msg = f"->{self.bed_name}|low volume:{volume}mL"
                mqttclient.publish(self.cfg.ALARM_TOPIC, msg)
                print("[报警] 已发送:", msg)
                self.has_alarmed = True
            except Exception as e:
                print("[网络] 报警发送失败:", str(e))
        # 声音报警
        try:
            self.buzzer.tone(
                self.cfg.BUZZER_FREQ,
                self.cfg.BUZZER_DURATION_MS
            )
        except Exception as e:
            print("[硬件] 蜂鸣器异常:", str(e))

    def _shutdown(self):
        """安全关闭系统"""
        self.is_running = False
        
        # 停止蜂鸣器
        self.buzzer.stop()
        
        # 清空显示屏
        self.display.clear()
        
        # 断开网络连接
        try:
            mqttclient.disconnect()
        except:
            pass
        print("[系统] 系统已关闭")
        sys.exit(0)

    def __recode_start(self):
        """输液开始记录"""
        try:
            start_time = datetime()
            msg = f"->{self.bed_name}|start time:{start_time}"
            mqttclient.publish(self.cfg.START_TOPIC, msg)
            print("[事件] 输液开始:", msg)
        except Exception as e:
            print("[网络] 输液开始发送失败:", str(e))           

    def __recode_end(self):
        """输液结束记录"""
        try:
            end_time = datetime()
            msg = f"->{self.bed_name}|end time:{end_time}"
            mqttclient.publish(self.cfg.END_TOPIC, msg)
            print("[事件] 输液结束:", msg)
        except Exception as e:
            print("[网络] 输液开始发送失败:", str(e)) 

    def run(self):
        """主运行循环"""
        try:
            # 打印设备唯一标识
            print(f"[系统] UUID：{self.uuid}")
            # 建立网络连接
            self.connect_network()
            self.__recode_start()  # 输液开始记录
            last_update = 0        # 上次刷新时间          
            # 主循环
            while self.is_running:
                current_time = time.ticks_ms()
                
                # 处理传感器信号
                self._handle_sensor(current_time)
                
                # 定时刷新界面
                if time.ticks_diff(current_time, last_update) >= self.cfg.DISPLAY_REFRESH_MS:
                    self._update_interface(current_time)
                    last_update = current_time
                
                # 释放CPU
                time.sleep_ms(10)
                
        except KeyboardInterrupt:
            print("[系统] 用户中断操作")
        except Exception as e:
            print("[系统] 严重错误:", str(e))
        finally:
            self._shutdown()

# --------------------------
# 模拟器（用于无硬件测试）
# --------------------------
class InfusionSimulator(InfusionMonitor):
    """输液模拟器（继承自主监控系统）"""
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.base_rate = 50     # 基准滴速（滴/分钟）
        self.next_drop_time = 0 # 下次滴液时间

    def _handle_sensor(self, current_time):
        """模拟滴液生成（替代真实传感器处理）"""
        if current_time >= self.next_drop_time:
            # 生成随机间隔（±20%波动）
            variance = random.uniform(0.8, 1.2)
            interval = (60000 // self.base_rate) * variance
            self.next_drop_time = current_time + int(interval)
            
            # 安全增加滴数
            max_drops = self.total_volume * self.drop_factor
            if self.drop_count < max_drops:
                self.drop_count += 1
                print(f"[模拟] 滴数：{self.drop_count}")

    def _update_interface(self, current_time):
        """增强界面更新（添加控制台输出）"""
        metrics = super()._update_interface(current_time)
        
        # 控制台打印监控数据
        print("\n==== 实时监控 =====")
        print(f"当前滴速: {metrics.drop_rate:.1f} 滴/分钟")
        print(f"已用时间: {metrics.elapsed_time}")
        print(f"剩余液量: {metrics.remaining_volume:.1f}mL")
        print(f"预计剩余: {metrics.remaining_time}")
        print("===================\n")      
        return metrics

# --------------------------
# 主程序入口
# --------------------------
if __name__ == "__main__":
    # 初始化配置
    config = AppConfig()
    
    try:
        # 创建真实传感器实例
        '''
        monitor = InfusionMonitor(
             cfg=config,
             bed_name="Bed-1",
             total_volume=250,
             drop_factor=20,
             alarm_threshold=10
        ) 
        '''
        
        # 创建模拟器实例
        monitor = InfusionSimulator(
            cfg=config,
            bed_name="SIM-1",  # 模拟床位号
            total_volume=2,      # 总液量30mL（便于快速测试）
            drop_factor=20,       # 标准滴系数
            alarm_threshold=1    # 25mL触发报警
        )
        # 启动监控系统
        monitor.run()    
    except Exception as e:
        print("[启动] 系统初始化失败:", str(e))
        sys.exit(1)
