"""
护士站中央监护系统（MicroPython版）
功能：多床位状态监控、报警提示、声光警示
硬件要求：行空板K10、有源蜂鸣器
"""

from educore import pin, wifi, mqttclient, oled, speaker, uuid
import time
import sys

# --------------------------
# 常量配置（与输液设备一致）
# --------------------------
class AppConfig:
    """应用程序全局配置"""

    # WiFi配置
    WIFI_SSID = "GOCLOUD"          # 无线网络名称
    WIFI_PASSWORD = "L2100Hbb"     # WiFi密码
    WIFI_TIMEOUT_MS = 50000        # 连接超时（毫秒）
    WIFI_MAX_RETRIES = 3           # 最大重试次数

    # MQTT配置
    MQTT_SERVER = "192.168.123.16"      # 消息代理地址
    MQTT_PORT = 1883                     # 服务端口
    MQTT_USER = "siot"                   # 认证用户名
    MQTT_PASSWORD = "dfrobot"            # 认证密码
    MQTT_MAX_RETRIES = 3                 # 连接重试次数
    ALARM_TOPIC = "siot/infusion_alarm"  # 报警主题
    START_TOPIC = "siot/infusion_start"  # 输液开始主题
    END_TOPIC = "siot/infusion_end"      # 输液结束主题

    # 硬件参数
    DISPLAY_REFRESH_MS = 1000         # 屏幕刷新间隔（毫秒）
    SPEAKER_PIN = 1                   # 蜂鸣器引脚
    BUZZER_FREQ = 2000                # 蜂鸣器频率（Hz）
    BUZZER_DURATION_MS = 500          # 蜂鸣持续时间（毫秒）

    # 显示参数
    SCREEN_WIDTH = 240
    SCREEN_HEIGHT = 320
    BED_COLUMNS = 2                   # 每行显示床位数
    BED_ROWS = 3                      #床位单元边距
    RECT_WIDTH = 90
    RECT_HEIGHT = 90
        
    # 报警坐标
    x_coordinates = [40, 150, 40, 150, 40, 150]
    y_coordinates = [80, 80, 180, 180, 280, 280]


# --------------------------
# 系统状态管理
# --------------------------
class NurseStation:
    def __init__(self, cfg):
        self.cfg = cfg
        self.alarm_beds = set()          # 报警床位集合
        self.last_alarm_time = 0         # 上次报警时间
        self.uuid = uuid()
        self.last_alarm_msg = None       # 记录上一次获取的报警消息
        self.last_start_msg = None       # 记录上一次获取的开始消息
        self.last_end_msg = None         # 记录上一次获取的结束消息
        # 初始化运行状态
        self._reset_state()
        # 初始化硬件
        self._init_hardware()

    def _reset_state(self):
        """重置所有运行时状态"""
        self.has_alarmed = False       # 报警状态标记
        self.is_running = True         # 运行状态标志

    def _init_hardware(self):
        """硬件初始化"""
        # OLED显示屏
        self.display = oled
        # 蜂鸣器
        self.buzzer = speaker(self.cfg.SPEAKER_PIN)
        self.buzzer.stop()

    def connect_network(self):
        """建立网络连接（WiFi+MQTT）"""
        self._connect_wifi()
        self._connect_mqtt()

    def _connect_wifi(self):
        """连接WiFi网络（带重试机制）"""
        for attempt in range(1, self.cfg.WIFI_MAX_RETRIES + 1):
            try:
                wifi.connect(
                    ssid=self.cfg.WIFI_SSID,
                    psd=self.cfg.WIFI_PASSWORD,
                    timeout=self.cfg.WIFI_TIMEOUT_MS
                )
                if wifi.status():
                    print("[网络] WiFi连接成功")
                    return
            except Exception as e:
                print(f"[网络] WiFi连接异常（尝试{attempt}/{self.cfg.WIFI_MAX_RETRIES}）: {str(e)}")
        raise RuntimeError("WiFi连接失败")

    def _connect_mqtt(self):
        """连接MQTT代理（带重试机制）"""
        for attempt in range(1, self.cfg.MQTT_MAX_RETRIES + 1):
            try:
                mqttclient.connect(
                    server=self.cfg.MQTT_SERVER,
                    port=self.cfg.MQTT_PORT,
                    client_id="NurseStation",
                    user=self.cfg.MQTT_USER,
                    psd=self.cfg.MQTT_PASSWORD
                )
                if mqttclient.connected():
                    print("[网络] MQTT连接成功")
                    return
            except Exception as e:
                print(f"[网络] MQTT连接异常（尝试{attempt}/{self.cfg.MQTT_MAX_RETRIES}）: {str(e)}")
        raise RuntimeError("MQTT连接失败")

    def _check_mqtt_alarm_message(self):
        try:
            # 主动获取指定主题消息
            msg = mqttclient.message(topic=self.cfg.ALARM_TOPIC)
            if msg is not None and msg != self.last_alarm_msg and msg:
                # 当有未处理且和上一次不同的消息时处理
                self._process_alarm(msg)
                self.last_alarm_msg = msg  # 更新上一次获取的报警消息
        except Exception as e:
            print("消息获取异常:", e)

    def _check_mqtt_start_message(self):
        try:
            # 主动获取指定主题消息
            msg = mqttclient.message(topic=self.cfg.START_TOPIC)
            if msg is not None and msg != self.last_start_msg and msg:
                # 当有未处理且和上一次不同的消息时处理
                self._process_start(msg)
                self.last_start_msg = msg  # 更新上一次获取的开始消息
        except Exception as e:
            print("消息获取异常:", e)

    def _check_mqtt_end_message(self):
        try:
            # 主动获取指定主题消息
            msg = mqttclient.message(topic=self.cfg.END_TOPIC)
            if msg is not None and msg != self.last_end_msg and msg:
                # 当有未处理且和上一次不同的消息时处理
                self._process_end(msg)
                self.last_end_msg = msg  # 更新上一次获取的消息
        except Exception as e:
            print("消息获取异常:", e)

    def _process_alarm(self, msg):
        """处理报警消息
        消息格式："Bed-1|low volume:0.95mL"
        """
        try:
            # 先按 | 分割消息
            parts = msg.split('|')
            # 提取床号部分，例如从 Bed-1 中提取 1
            bed_name = parts[0]
            alarm_msg = parts[1]
            bed_number = int(bed_name.split('-')[1])
            print(f"[报警] 低液量：{bed_name} {alarm_msg}")
            self.alarm_beds.add(bed_name)
            self.last_alarm_time = time.ticks_ms()
            self.alarm_flashing = True  # 开启闪烁标志
            # 文字闪烁逻辑
            for _ in range(5):  # 闪烁5次，可根据需要调整
                self._trigger_buzzer()
                self.display.print("Low Alarm", self.cfg.x_coordinates[bed_number - 1], self.cfg.y_coordinates[bed_number - 1])
                time.sleep(0.5)
                self.display.print("         ", self.cfg.x_coordinates[bed_number - 1], self.cfg.y_coordinates[bed_number - 1])
                time.sleep(0.5)
            self.alarm_flashing = False  # 关闭闪烁标志
        except ValueError as e:
            print(f"消息解析错误，无法将床号转换为整数: {e}")
        except IndexError as e:
            print(f"消息格式错误，无法正确分割消息: {e}")
        except Exception as e:
            print(f"消息处理异常: {e}")

    def _process_start(self, msg):
        """处理报警消息
        消息格式："Bed-1|start time:2025-02-06 12:28:15"
        """
        try:
            bed_name, start_time = msg.split('|')
            print(f"[事件] {bed_name} {start_time}")
            bed_number = int(bed_name.split('-')[1])  # 分割获取床号
            self.display.print(bed_name, self.cfg.x_coordinates[bed_number - 1]+10, self.cfg.y_coordinates[bed_number - 1] - 50)
        except Exception as e:
            print("消息解析失败:", e)

    def _process_end(self, msg):
        """处理报警消息
        消息格式："Bed-1|end time:2025-02-06 12:28:15"
        """
        try:
            bed_name, end_time = msg.split('|')
            print(f"[事件] {bed_name} {end_time}")
            bed_number = int(bed_name.split('-')[1])  # 分割获取床号
            self.display.print("      ", self.cfg.x_coordinates[bed_number - 1]+10, self.cfg.y_coordinates[bed_number - 1] - 50)
        except Exception as e:
            print("消息解析失败:", e)

    def _on_alarm(self, topic, msg):
        """报警消息回调"""
        try:
            bed_name, alarm_msg = msg.split('|')
            print(f"[网络] {topic}：{bed_name} {alarm_msg}")

            # 更新报警状态
            self.alarm_beds.add(bed_name)
            self.last_alarm_time = time.ticks_ms()

            # 触发声音报警
            self._trigger_buzzer()
        except:
            print("无效报警消息:", msg)

    def _trigger_buzzer(self):
        """控制蜂鸣器（非阻塞）"""
        try:
            self.buzzer.tone(self.cfg.BUZZER_FREQ, self.cfg.BUZZER_DURATION_MS)
        except Exception as e:
            print("蜂鸣器异常:", e)

    def _draw_bed(self, x, y, w, h, bed_name):
        """绘制单个床位状态"""
        # 绘制矩形的左边框
        for i in range(1, h - 1):
            try:
                self.display.print('|', x, y + i)
            except Exception as e:
                print(f"Error printing left border at ({x}, {y + i}): {e}")

        # 绘制矩形的右边框
        for i in range(1, h - 1):
            try:
                self.display.print('|', x + w - 1, y + i)
            except Exception as e:
                print(f"Error printing right border at ({x + w - 1}, {y + i}): {e}")
        # 绘制矩形的上边框
        self.display.print('--------------', x + 5, y - 2)

        # 绘制矩形的下边框
        self.display.print('--------------', x + 5, y + h + 1)

        '''
        # 计算文本的位置，将文本居中显示在矩形内
        text_x = x + w // 2 - len(bed_name) // 2
        text_y = y + h // 2
        text_width = len(bed_name) * 5  # 估算字符宽度
        text_x = x + (w - text_width)//2
        text_y = y + (h - 16)//4  # 字体高度约16px
        # 居中显示床号
        self.display.print(bed_name, text_x, text_y)
        '''

    def _draw_multiple_beds(self, columns, rows, rect_width, rect_height):
        """绘制多个床位状态"""
        screen_width = self.cfg.SCREEN_WIDTH
        screen_height = self.cfg.SCREEN_HEIGHT

        # 计算横向和纵向总的占用空间（包含矩形框和间隔）
        total_width = columns * rect_width
        total_height = rows * rect_height

        # 计算横向和纵向的间隔总数
        horizontal_gap_count = columns - 1
        vertical_gap_count = rows - 1

        # 计算横向和纵向的总间隔长度
        available_horizontal_space = screen_width - total_width
        available_vertical_space = screen_height - total_height

        # 计算横向和纵向的间隔大小
        horizontal_gap = available_horizontal_space // (horizontal_gap_count + 2)
        vertical_gap = available_vertical_space // (vertical_gap_count + 2)

        # 计算起始的 x 和 y 坐标，使整个矩形框组居中
        start_x = horizontal_gap
        start_y = vertical_gap

        count = 1
        for row in range(rows):
            for col in range(columns):
                # 计算当前矩形的左上角坐标
                x = start_x + col * (rect_width + horizontal_gap)
                y = start_y + row * (rect_height + vertical_gap)
                text = f"BED-{count}"
                self._draw_bed(x, y, rect_width, rect_height, text)
                count += 1

    def _display_all_beds(self):
        """显示所有床位"""
        self._draw_multiple_beds(self.cfg.BED_COLUMNS, self.cfg.BED_ROWS, self.cfg.RECT_WIDTH, self.cfg.RECT_HEIGHT)

    def run(self):
        """主运行循环"""
        try:
            # 打印设备唯一标识
            print(f"[系统] UUID：{self.uuid}")
            # 建立网络连接
            self.connect_network()
            # 显示所有床位
            self._display_all_beds()
            # 打印护士站初始化完成
            print(f"[系统] 护士站初始化完成")
            while self.is_running:
                # 检查输液开始消息
                self._check_mqtt_start_message()
                # 检查输液报警消息
                self._check_mqtt_alarm_message()
                # 检查输液结束消息
                self._check_mqtt_end_message()
                # 自动清除过期报警（5秒后自动清除）
                if time.ticks_diff(time.ticks_ms(), self.last_alarm_time) > 5000:
                    self.alarm_beds.clear()
                time.sleep(0.5)
        except KeyboardInterrupt:
            self._shutdown()

    def _shutdown(self):
        """安全关闭系统"""
        self.is_running = False

        # 停止蜂鸣器
        self.buzzer.stop()

        # 清空显示屏
        self.display.clear()

        # 断开网络连接
        try:
            mqttclient.disconnect()
        except:
            pass
        print("[系统] 系统已关闭")
        sys.exit(0)


# --------------------------
# 主程序入口
# --------------------------
if __name__ == "__main__":
    config = AppConfig()
    station = NurseStation(config)
    station.run()